/*
 * Copyright (c) 2001 by Cliff Berg.
 * This code may be used and distributed for research for purposes.
 * No warrantee of any kind is given or implied.
 */

package mc.myupdateable;

import java.awt.Frame;
import java.awt.Button;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.net.URLClassLoader;
import java.net.URL;
import java.rmi.RMISecurityManager;
import java.io.File;

import mc.myupdateable.*;
import mc.rmi.*;

public class Test
{
	static MCRuntime runtime;

	static Frame frame;

	static LocalReference serviceLocalRef;

	public static void main(String[] args) throws Exception
	{
		System.out.println("Begin Test");
		System.out.println("java.home=" + System.getProperty("java.home"));
		System.out.println("user.home=" + System.getProperty("user.home"));

		System.setSecurityManager(new RMISecurityManager());

		if (args.length < 1)
		{
			System.out.println("Usage: java mc.myupdateable.Test <cachedir>");
			return;
		}

		String cacheDirString = args[0];
		File cacheDir = new File(cacheDirString);

		runtime = new MCRuntime(cacheDir);

		frame = new Frame();
		frame.setSize(150, 75);

		//
		// For an application to get services, it needs to be a service
		// with a createRequestor() method. So here I create a bootstrap
		// service that I can use to create a Requestor to get other services.

		Service bootstrapSvc = new Service()
		{
			public int[] getServiceVersionNo()
			{
				return new int[] { 1 };
			}

			public Requestor createRequestor()
			{
				return null; // because this will never be called.
			}

			public Requestor createRequestor(final int[] serviceVersionNoRequestorHas,
				final Class serviceInterfaceRequestorHas)
			{
				return new Requestor()
				{
					public int[] getServiceVersionNoRequestorHas()
					{
						return serviceVersionNoRequestorHas;
					}

					public Class getServiceInterfaceRequestorHas()
					{
						return serviceInterfaceRequestorHas;
					}
				};
			}

			public Configurator getConfigurator(ConfigRuntime rt)
			{
				return null; // because this will never be called.
			}

			public void init()
			{
			}

			public void start()
			{
			}

			public void stop()
			{
			}

			public void deactivate()
			{
			}

			public ServiceState getState()
			{
				return null; // because this will never be called.
			}
		};

		//
		// Create an initial instance of MyUpdateable. Load it explicitly
		// from a remote location, using a new URLClassLoader.
		// Make sure that MyUpdateable is not in the classpath, because the
		// URLClassLoader will first delegate to the system class loader
		// (as it should), and we want to force the URLClassLoader to
		// actually do the load. (To do this, the MyUpdateable implementation
		// classes must be at the URL that the URLClassLoader will search.)

		System.out.println("Getting service...");
		serviceLocalRef = runtime.getService
		(
			"mc.myupdateable.MyUpdateable",
			ServiceProvider.ServiceName,
			"localhost",
			bootstrapSvc.createRequestor(null, null),
			false  // if we already have it, do not check for an update.
		);

		synchronized (serviceLocalRef)
		{
			Service s = (Service)(serviceLocalRef.get(null));

			s.start();

			System.out.println("Version=" + (s.createRequestor().getServiceVersionNoRequestorHas())[0]);
		}

		//
		// Instantiate a button to allow the user to initiate an update at
		// any time.

		Button b = new Button("Update");
		b.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					//
					// Spawn a new thread to initiate the update operation. The
					// reason is that the update operation blocks until it is done,
					// and since this action is occurring in the UI thread,
					// we don't want to block the UI thread.

					(
						new Thread
						(
							new Runnable()
							{
								public void run()
								{
									try
									{
										synchronized (serviceLocalRef)
										{
											Service s = (Service)(serviceLocalRef.get(null));

											if (s instanceof UpdateableService)
											{
												UpdateableService u = (UpdateableService)s;
												LocalReference r = runtime.update(u);
												Service newS = (Service)(r.get(null));
												s = newS;
												s.start();

												System.out.println("Updated.");
												System.out.println("Version=" +
													s.getServiceVersionNo()[0]);
											}
											else
											{
												System.out.println(
													"s is not updateable");
											}
										}
									}
									catch (Exception ex)
									{
										System.out.println("Update failed:");
										ex.printStackTrace();
										return;
									}
								}
							}
						)
					)
					.start();
				}
			}
		);

		frame.add(b);

		frame.show();
	}
}

