/*
 * Copyright (c) 2001 by Cliff Berg.
 * This code may be used and distributed for research for purposes.
 * No warrantee of any kind is given or implied.
 */

package mc.rmi;

import java.rmi.RemoteException;

public interface ConfigRuntime
{
	// ------------------------------------------------------------
	// Service lookup, retrieval, and installation.

	/**
	 * Find and install the specified service within this runtime.
	 * If host is null, it is expected that the application has
	 * already been obtained, and the client is merely requesting
	 * a reference to it.
	 */
	public LocalReference getService
	(
		String svcInterfaceName,
		String svcProviderName,
		String host,
		Requestor requestor,
		boolean update  // if true, an update will be checked for.
	)
	throws
		NoUpdateAvailable,
	 	ConfigurationAlert,
	 	ConfigurationFailure,
	 	ConnectionFailure,
	 	ServiceNotFound,
	 	InsufficientResources,
	 	InitializationFailure,
		RemoteException;

	/**
	 * Supplies a null value for host. This implies that the
	 * service is already installed, and we merely want to
	 * return a reference to it.
	 */
	public LocalReference getService
	(
		String svcInterfaceName,
		String svcProviderName,
		Requestor requestor,
		boolean update  // if true, an update will be checked for.
	)
	throws
		NoUpdateAvailable,
	 	ConfigurationAlert,
	 	ConfigurationFailure,
	 	ConnectionFailure,
	 	ServiceNotFound,
	 	InsufficientResources,
	 	InitializationFailure,
		RemoteException;

	// ------------------------------------------------------------
	// Service update.

	/**
	 * Allows an application to register for notification when an update
	 * for a service becomes available. However, notification is not
	 * guaranteed and is not synchronous.
	 */
	public void listenForUpdate
	(
		Service svc
	)
	throws
		Exception;

	/**
	 * Attempt to obtain an updated version of the specified updateable
	 * object and return a new local reference to it. If the service
	 * is not installed, a runtime exception is thrown. The instance being
	 * replaced is removed from the runtime and listeners are notified.
	 * However, if u is current and does not need to be updated, then
	 * a CallerIsCurrent exception is thrown.
	 */
	public LocalReference update(UpdateableService u)
	throws
		CallerIsCurrent,
		NoUpdateAvailable,
	 	ConfigurationAlert,
	 	ConfigurationFailure,
	 	ConnectionFailure,
	 	ServiceNotFound,
	 	UpdateFailure,
	 	InsufficientResources,
	 	InitializationFailure,
		RemoteException;

	/**
	 * Uninstall the specified service.
	 */
	public LocalReference uninstallService
	(
		String svcInterfaceName
	)
	throws
	 	ServiceNotFound;

	// ------------------------------------------------------------
	// Runtime lifecycle methods.

	/**
	 * Load all installed services. Do not attempt to return
	 * them to initialize or start them.
	 */
	public void start();

	/**
	 * Gracefully terminate all services. This includes:
	 *
	 *   Calling stop() on each service.
	 *   Closing all managed resources, once all stop() methods
	 *      have completed or timed out.
	 */
	public void shutdown();

	// ------------------------------------------------------------
	// Methods for obtaining managed resources.

	/**
	 * Obtain a thread for use by a service. Services that need an
	 * execution thread (beyond the UI thread) should obtain one
	 * via this method.
	 */
	public ManagedThread getThread();
}
