/*
 * Copyright (c) 2001 by Cliff Berg.
 * This code may be used and distributed for research for purposes.
 * No warrantee of any kind is given or implied.
 */

package mc.rmi;

import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.lang.ref.WeakReference;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;

public abstract class LocalReferenceBase implements LocalReference
{
	private transient ConfigRuntime runtime; // don't serialize.
	private transient Object referent; // don't serialize.
	private transient SortedSet listeners; // of weak references
		// to NullificationListener.

	protected LocalReferenceBase()
	{
		this.listeners = new TreeSet();
	}

	public LocalReferenceBase(ConfigRuntime r, Object localState)
	{
		this();
		this.runtime = r;
		this.referent = localState;
	}

	/**
	 * Reclaim the reference. Users of the reference are expected to
	 * synchronize on this local reference object during the time
	 * they hold a reference to the referent; and to have registered a
	 * listener for notification when reclamation occurs. This method
	 * should notify those listeners.
	 * Note: this should be implemented with a timeout.
	 */
	public synchronized void clear()
	throws
		ReferenceAccessTimeout
	{
		referent = null;

		notifyListeners(false);
	}

	/**
	 * Obtain a reference to the actual (managed) object.
	 * A notification listenener may be registered at this time.
	 * Successive calls to get() may specify the same listener.
	 * However, a listener should only listen for notifications for
	 * a single local reference object.
	 */
	public synchronized Object get
	(
		NullificationListener listener
	)
	throws
		ReferentIsNull
	{
		if (referent == null)
		{
			throw new ReferentIsNull();
		}

		// Create a weak reference for the listener.

		WeakReference wr = new WeakReference(listener);

		// Add listener (non-redundantly) to the notification list.

		if (listener != null) listeners.add(wr);

		return referent;
	}

	/**
	 * Check if the referent is null.
	 */
	public synchronized boolean isNull()
	{
		return referent == null;
	}

	/**
	 * Set the value of the reference (ie. the value returned by get()).
	 * If the reference already contains a value, a runtime exception is
	 * thrown.
	 *
	 * If the runtime times out trying to access this, a
	 * runtime exception is also thrown. (Not implemented.)
	 */
	public synchronized void set(Object value)
	{
		if (referent != null)
		{
			throw new RuntimeException(
				"Attempt to overwrite a local reference's value.");
		}

		if (value == null)
		{
			throw new RuntimeException(
				"Attempt to set local reference to a null value.");
		}

		referent = value;

		notifyListeners(true);
	}

	// ------------------------------------------------------------
	// Protected methods.

	/**
	 * Notify all listeners that this local reference has been
	 * modified. If set is true, call referenceHasBeenSet() on each
	 * listener; otherwise, call referenceHasBeenCleared().
	 */
	protected void notifyListeners(boolean set)
	{
		Iterator i = listeners.iterator();
		try
		{
			WeakReference wr = (WeakReference)(i.next());
			NullificationListener n = (NullificationListener)(wr.get());
			if (n != null)
			{
				LocalReference lref = n.getListenersLocalReference();

				if ((lref == null) || (lref.isNull()))				{
					// The listener is no longer present.
					// Dispose of the weak reference to the listener.

					wr.clear();
					listeners.remove(wr);
				}
				else
				{
					// Notify the listener that this local reference
					// has been set or cleared.

					if (set)
					{
						n.referenceHasBeenSet();
					}
					else
					{
						n.referenceHasBeenCleared();
					}
				}
			}
		}
		catch (NoSuchElementException ex)
		{
			// Done.
		}
	}

	// ------------------------------------------------------------
	// Serialization override methods.

	protected void setRuntime(ConfigRuntime r)
	{
		this.runtime = r;
	}

	private void writeObject(ObjectOutputStream stream)
	throws
		IOException
	{
		stream.defaultWriteObject();
	}

	private void readObject(ObjectInputStream stream)
	throws
		IOException,
		ClassNotFoundException
	{
		stream.defaultReadObject();

		this.listeners = new TreeSet();
	}
}
