/*
 * Copyright (c) 2001 by Cliff Berg.
 * This code may be used and distributed for research for purposes.
 * No warrantee of any kind is given or implied.
 */

package mc.rmi;

import java.rmi.RemoteException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;

import java.net.URLClassLoader; // I hacked this.
import java.net.NullClassLoader; // I added this.

public class MCRuntime implements ConfigRuntime
{
	private static final String RemoteUpdaterClassName =
		"mc.rmi.RemoteUpdater";
	private static final String ServiceListFilename =
		"ServiceList.ser";

	// ------------------------------------------------------------
	// Private instance state.

	private File rootDir;
	private ServiceList svcList;

	// ------------------------------------------------------------
	// Constructors.

	public MCRuntime(File rootDir)
	{
		this.rootDir = rootDir;

		RuntimeRegistry.setRuntime(this);

		try
		{
			this.svcList = new ServiceList(new File(rootDir,
				ServiceListFilename));
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
			System.out.println("Service list could not be loaded");
		}

		start();
	}

	// ------------------------------------------------------------
	// From ConfigRuntime.

	/**
	 * Find and install the specified service within this runtime.
	 * The host argument is the name service host.
	 * If host is null, it is expected that the application has
	 * already been obtained, and the client is merely requesting
	 * a reference to it.
	 */
	public LocalReference getService
	(
		String svcInterfaceName,
		String svcProviderName,
		String host,
		Requestor requestor,
		boolean update
	)
	throws
		NoUpdateAvailable,
	 	ConfigurationAlert,
	 	ConfigurationFailure,
	 	ConnectionFailure,
	 	ServiceNotFound,
	 	InsufficientResources,
	 	InitializationFailure,
		RemoteException
	{
		//
		// First check if we have it.
		// If not, request it from the specified host. If host is null
		// assume the local host.
		//

		// Check our service list to see if we have it.

		ServiceDescriptor sd = svcList.findService(svcInterfaceName, true);

		if (sd != null)
		{
			// We already have it.

			if (update)
			{
				Object o;
				try
				{
					o = sd.lref.get(null);
				}
				catch (ReferentIsNull ex)
				{
					ex.printStackTrace();
					throw new ConfigurationFailure(ex,
						"Local reference to " + svcInterfaceName +
						" is null.");
				}

				if (! (o instanceof UpdateableService))
				{
					System.out.println(
						"Service " + svcInterfaceName +
						" is not updateable; using installed version.");

					return sd.lref;
				}

				UpdateableService us = (UpdateableService)o;
				try
				{
					return this.update(us);
				}
				catch (Exception ex)
				{
					// An error during update attempt; use the version we have.
					return sd.lref;
				}
			}

			return sd.lref;
		}

		//
		// Obtain a new instance of the service. This configures
		// (initializes) the service, but does not start it.

		Service newService;
		try
		{
			newService = getServiceFromProvider(svcInterfaceName,
				svcProviderName, host, requestor);
		}
		catch (CallerIsCurrent ex)
		{
			ex.printStackTrace();
			throw new ConfigurationFailure(ex,
				"Unexpected failure");
		}
		catch (UpdateFailure ex2)
		{
			ex2.printStackTrace();
			throw new ConfigurationFailure(ex2,
				"Unexpected failure");
		}

		//....Note: there is a synchronization issue re. starting
		// the service and making it available to other threads.

		//
		// Install the service.

		sd = svcList.addService(newService, host, svcProviderName);

		//
		// Return the new object that replaces the old one.

		return sd.lref;
	}

	/**
	 * Supplies a null value for host. This implies that the
	 * service is already installed, and we merely want to
	 * return a reference to it.
	 */
	public LocalReference getService
	(
		String svcInterfaceName,
		String svcProviderName,
		Requestor requestor,
		boolean update
	)
	throws
		NoUpdateAvailable,
	 	ConfigurationAlert,
	 	ConfigurationFailure,
	 	ConnectionFailure,
	 	ServiceNotFound,
	 	InsufficientResources,
	 	InitializationFailure,
		RemoteException
	{
		return getService(svcInterfaceName, svcProviderName, null, update);
	}

	/**
	 * Allows an application to register for notification when an update
	 * for a service becomes available. However, notification is not
	 * guaranteed and is not synchronous.
	 */
	public void listenForUpdate
	(
		Service svc
	)
	throws
		Exception
	{
		throw new RuntimeException("Not implemented");
	}

	/**
	 * Attempt to obtain an updated version of the specified updateable
	 * object and return a new local reference to it. If the service
	 * is not installed, a runtime exception is thrown. The instance being
	 * replaced is removed from the runtime and listeners are notified.
	 * However, if u is current and does not need to be updated, then
	 * a CallerIsCurrent exception is thrown.
	 */
	public LocalReference update
	(
		UpdateableService u
	)
	throws
		CallerIsCurrent,
		NoUpdateAvailable,
	 	ConfigurationAlert,
	 	ConfigurationFailure,
	 	ConnectionFailure,
	 	ServiceNotFound,
	 	UpdateFailure,
	 	InsufficientResources,
	 	InitializationFailure,
		RemoteException
	{
		//
		// Identify the service.

		Class[] interfaces = u.getClass().getInterfaces();
		if (interfaces.length == 0)
		{
			throw new UpdateFailure(
				"Unexpected: the service does not implement an interface.");
		}

		String svcInterfaceName = interfaces[0].getName();

		//
		// Check our service list to see if we have it.

		ServiceDescriptor sd = svcList.findService(svcInterfaceName, true);

		if (sd == null)
		{
			throw new RuntimeException(
				"Could not find service descriptor for " + svcInterfaceName);
		}

		String svcProviderName = sd.svcProviderName;
		String host = sd.host;

		//
		// Create a requestor to carry information about the client.

		Requestor requestor = u.createRequestor();

		//
		// Deactivate the previous version of the service.

		LocalReference lref;

		try
		{
			// Tell the service to shut down.

			u.deactivate();
			System.out.println("Service deactivated");

			// Disconnect the service from the runtime.
			// This may block indefinitely if it is not implemented with
			// a timeout.

			lref = svcList.removeService(u);
			sd = null;
		}
		catch (ReferenceAccessTimeout ex)
		{
			throw new UpdateFailure(ex);
		}
		catch (Exception ex2)
		{
			throw new UpdateFailure(ex2);
		}

		//
		// Now obtain a new instance of the service. This configures
		// (initializes) the service, but does not start it.

		Service newService = getServiceFromProvider(svcInterfaceName,
			svcProviderName, host, requestor);

		//
		// Start the new version. Reuse the local reference that the service
		// had before, so that holders of the local reference can
		// continue to use the local reference to access the service.

		sd = svcList.addService(newService, lref, host, svcProviderName);

		//
		// Return the new object that replaces the old one.

		return sd.lref;
	}

	/**
	 * Uninstall the specified service.
	 */
	public LocalReference uninstallService
	(
		String svcInterfaceName
	)
	throws
	 	ServiceNotFound
	{
		throw new RuntimeException("Not implemented");
	}

	/**
	 * Load all installed services. Do not attempt to return
	 * them to initialize or start them.
	 */
	public void start()
	{
		try
		{
			svcList.load(false);
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
			System.out.println("Error loading service list.");
		}
	}

	/**
	 * Gracefully terminate all services. This includes:
	 *
	 *   Calling stop() on each service.
	 *   Closing all managed resources, once all stop() methods
	 *      have completed or timed out.
	 */
	public void shutdown()
	{
		// For now, do nothing.
	}

	/**
	 * Obtain a thread for use by a service. Services that need an
	 * execution thread (beyond the UI thread) should obtain one
	 * via this method.
	 */
	public ManagedThread getThread()
	{
		throw new RuntimeException("ManagedThread not implemented");
	}

	// ------------------------------------------------------------
	// Protected and private definitions.

	/**
	 * Retrieve the specified service from the specified host,
	 * and return it.
	 */
	protected Service getServiceFromProvider
	(
		String svcInterfaceName,
		String svcProviderName,
		String host, // name service host.
		Requestor requestor
	)
	throws
	 	CallerIsCurrent,
		NoUpdateAvailable,
	 	ConfigurationAlert,
	 	ConfigurationFailure,
	 	ConnectionFailure,
	 	ServiceNotFound,
	 	UpdateFailure,
	 	InsufficientResources,
		RemoteException
	{
		//
		// Instantiate a remote class loader.

		ClassLoader classLoader = getNewClassLoader();  // This is
			// a direct child of a root class loader, to ensure that
			// the class is not already loaded by any parent class loader.

		LoaderThread t = new LoaderThread(classLoader, svcInterfaceName,
			svcProviderName, host, requestor);

		t.start();

		try
		{
			t.join();  // Wait for it to finish. Note: I should add
				// a timeout here.
		}
		catch (InterruptedException ex)
		{
			throw new UpdateFailure(ex, "Update was interrupted");
		}

		return t.getService();
	}

	/**
	 * Create a new class loader for a service. The class loader will
	 * be an instance of NullClassLoader, which provides a caching
	 * mechanism for child class laoders to use.
	 */
	protected ClassLoader getNewClassLoader()
	throws
		InsufficientResources
	{
		return new NullClassLoader(getNewCacheDir());
	}

	/**
	 * Create a new directory for storing a service's classes and
	 * resources.
	 */
	protected File getNewCacheDir()
	throws
		InsufficientResources
	{
		File d = new File(rootDir, makeUniqueName());
		if (! d.mkdir())
		{
			throw new InsufficientResources();
		}

		System.out.println("Cache directory created: " + d);

		return d;
	}

	private static int uniqueNumber = 0;

	protected static String makeUniqueName()
	{
		uniqueNumber++;
		Date d = new Date();
		long t = d.getTime();
		String s = "_" + t + "_" + uniqueNumber;
		return s;
	}

	/**
	 * Implements all the functionality required of the service list.
	 * The act of installing a service consists essentially of creating
	 * a service list entry. Once the entry is created, the service
	 * is considered to be "installed". There is only one service
	 * list for a runtime.
	 */
	private class ServiceList
	{
		private File file;  // Where the list is stored.
		private Vector services;  // List of service descriptors.

		public ServiceList(File f)
		throws
			IOException
		{
			this.file = f;
		}

		/**
		 * Add the specified service to the list. If the service is
		 * already in the list, simply return.
		 */
		public ServiceDescriptor addService(Service service, String host,
			String svcProviderName)
		throws
			InsufficientResources,
			InitializationFailure
		{
			return addService(service, null, host, svcProviderName);
		}

		/**
		 * Add the specified service to the list. If the service is
		 * already in the list, simply return.
		 * If the specified local reference is not null, reuse it; in that case,
		 * its referent is assumed to be (must be) null.
		 */
		public ServiceDescriptor addService(Service service, LocalReference lref,
			String host, String svcProviderName)
		throws
			InsufficientResources,
			InitializationFailure
		{
			if ((lref != null) && (! lref.isNull()))
			{
				throw new RuntimeException("Referent is not null");
			}

			//
			// Get the service's name (same as its full classname).

			Class svcClass = service.getClass();
			Class[] intfcs = svcClass.getInterfaces();
			if (intfcs.length == 0)
			{
				throw new RuntimeException(
					"Unexptected: service " + service.getClass().getName() +
					" does not have an interface.");
			}

			String svcInterfaceName = intfcs[0].getName();

			//
			// Check if we already have the service in our list.

			ServiceDescriptor sd = findService(svcInterfaceName, false);
			if (sd != null) return sd;

			//
			// We don't have a descriptor for it. We need to create
			// one. This is the essence of "installing" the service.
			//

			//
			// Create or reuse the local reference, as required.

			if (lref == null)
			{
				lref = new MCLocalReference(MCRuntime.this, service);
			}
			else
			{
				lref.set(service);
			}

			//
			// Determine the directory that was used to cache the service's
			// classes. I am assuming that the service was loaded with a
			// URLClassLoader (my modified version).

			ClassLoader cl = service.getClass().getClassLoader();
			if (! (cl instanceof URLClassLoader))
			{
				throw new RuntimeException(
					"The class loader of " + service.getClass().getName() +
					" is not a URLClassLoader.");
			}

			URLClassLoader ucl = (URLClassLoader)cl;
			File cacheDir = ucl.getClassCacheDir();

			//
			// Create a service descriptor for the service.

			ServiceDescriptor sdNew = new ServiceDescriptor(host,
				svcInterfaceName, svcProviderName, svcClass.getName(),
				lref, cacheDir);

			//
			// Persistify the service descriptor.

			try
			{
				save();
			}
			catch (IOException ex)
			{
				ex.printStackTrace();
				throw new InsufficientResources(ex);
			}

			//
			// Load the service's persistent state, if any.

			try
			{
				service.init();
			}
			catch (Throwable t)
			{
				throw new InitializationFailure(t, lref);
			}

			//
			// Publish the service.

			services.add(sdNew); // After this, it is visible to other threads.

			return sdNew;
		}

		/**
		 * Clear the service's local reference and remove the
		 * service from the list. Return the local reference to the
		 * service being removed.
		 * If the service is not in the list, simply return null.
		 */
		public LocalReference removeService(Service service)
		throws
			ReferenceAccessTimeout
		{
			Class svcClass = service.getClass();
			Class[] intfcs = svcClass.getInterfaces();
			if (intfcs.length == 0)
			{
				throw new RuntimeException(
					"Unexptected: service " + service.getClass().getName() +
					" does not have an interface.");
			}

			String svcInterfaceName = intfcs[0].getName();
			ServiceDescriptor sd;
			try
			{
				sd = findService(svcInterfaceName, false);
			}
			catch (InitializationFailure ex)
			{
				ex.printStackTrace();
				throw new RuntimeException("Should not happen");
			}

			if (sd == null) return null;

			services.remove(sd);

			// Unpersistify the service descriptor.

			try
			{
				save();
			}
			catch (IOException ex)
			{
				ex.printStackTrace();
			}

			sd.lref.clear();

			// Note:
			// The services that get notified as a result of the above
			// statement may want to register to be notified when
			// the service becomes available again. This registration
			// need not be persistent, because those services should
			// also register their interest when they initialize.
			// Notification is performed by the local reference when
			// it is set to a new value.

			//
			// Delete the service's cached classes.
			// Note: I should really move them in case the update fails and
			// they need to be restored, but for now just delete them.

			File cacheDir = sd.getClassCacheDir();
			if ((cacheDir != null) && (! cacheDir.delete()))
			{
				System.out.println("Error deleting " + cacheDir);
			}

			return sd.lref;
		}

		/**
		 * Search for the specified service. If the service is not
		 * found, return null. If the service is found (i.e. is installed),
		 * but is not loaded, create a class loader for it and
		 * intantiate the service.
		 *
		 * A service returned by this method is assumed to be in an initialized
		 * state.
		 */
		public ServiceDescriptor findService
		(
			String svcInterfaceName,
			boolean load // if service is installed but not loaded, load it.
		)
		throws
			InitializationFailure
		{
			ServiceDescriptor sd = findServiceDescriptor(svcInterfaceName);

			if (sd == null) return null;

			//
			// We have it (it is installed.) Now load it if required.

			if (load)
			{
				LocalReference lref = sd.getLocalReference();

				// Load it if it is not loaded.

				if (lref.isNull())
				{
					// Create a new class loader, using the directory
					// that contains the service's classes.

					File d = sd.getClassCacheDir();
					ClassLoader cl = new NullClassLoader(d);

					// Explicitly load the implementation class, using
					// the new class loader.

					String svcClassName = sd.getClassName();
					Class c;
					try
					{
						c = cl.loadClass(svcClassName);
					}
					catch (ClassNotFoundException ex)
					{
						throw new Error(
							"Class " + svcClassName + " is missing.");
					}

					// Instantiate the service class.

					Service service;
					try
					{
						service = (Service)(c.newInstance());
					}
					catch (Exception ex)
					{
						throw new Error(
							"Unable to instantiate " + svcClassName);
					}

					lref.set(service);

					//
					// Restore service's persistent state.

					try
					{
						service.init();
					}
					catch (Throwable t)
					{
						throw new InitializationFailure(t, lref);
					}
				}
			}

			return sd;
		}

		/**
		 * Retrieve the service descriptor for the specified service.
		 * If the service is not installed (i.e. it has no descriptor),
		 * return null.
		 */
		protected ServiceDescriptor findServiceDescriptor
		(
			String svcInterfaceName
		)
		{
			for (int i = 0; i < services.size(); i++)
			{
				ServiceDescriptor sd = (ServiceDescriptor)(services.elementAt(i));

				if (sd.svcInterfaceName.equals(svcInterfaceName))
				{
					return sd;
				}
			}

			return null;
		}

		/**
		 * Write this service list to its file location.
		 */
		public void save()
		throws
			IOException
		{
			FileOutputStream fos = new FileOutputStream(file);
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			oos.writeObject(services);
			oos.close();
		}

		/**
		 * Load this service list from its file location. If replace is
		 * true, this method first checks if this list is empty, and if
		 * not, throws a runtime exception. If replace is false, this
		 * method does not check, and any existing items in this list
		 * are replaced with those that are loaded.
		 *
		 * Services are not automatically loaded by this method.
		 * Instead, all service descriptor attributes are reloaded
		 * and properly set; but the service's local reference has an
		 * empty value (lref.get() returns null). A service is loaded
		 * and initialized the next time it is requested via the
		 * findService(...) method.
		 */
		public void load(boolean replace)
		throws
			IOException
		{
			if ((! replace) && (services != null) && (services.size() > 0))
			{
				throw new RuntimeException(
					"Attempt to overwrite service list.");
			}

			FileInputStream fis;
			try
			{
				fis = new FileInputStream(file);
			}
			catch (FileNotFoundException ex)
			{
				services = new Vector();
				return;
			}

			ObjectInputStream ois = new ObjectInputStream(fis);
			Vector v;
			try
			{
				v = (Vector)(ois.readObject());
			}
			catch (ClassNotFoundException ex)
			{
				ex.printStackTrace();
				throw new Error("Could not read object");
			}

			ois.close();
			services = v;
		}
	}

	private class ServiceDescriptor
	{
		private ServiceDescriptor() {}

		public ServiceDescriptor(String host, String svcIntName,
			String svcProvName, String svcClassName, LocalReference lr,
			File cacheDir)
		{
			this.host = host;
			this.svcInterfaceName = svcIntName;
			this.svcProviderName = svcProvName;
			this.svcImplClassName = svcClassName;
			this.lref = lr;
			this.cacheDir = cacheDir;
		}

		public String getHost() { return host; }
		public String getProviderName() { return svcProviderName; }
		public String getClassName() { return svcImplClassName; }
		public LocalReference getLocalReference() { return lref; }
		public File getClassCacheDir() { return cacheDir; }

		// The service descriptor owns all these objects. All are
		// therefore serialized when the descriptor is serialized.
		// An exception is the service's local reference (lref):
		// its fields are transient and therefore will not be
		// serialized, and so when the descriptor is reloaded,
		// these fields will be null. The runtime must be able
		// to fill these fields as needed (lazily). The fields
		// include:
		//
		//		ConfigRuntime runtime
		//		Object referent
		//		SortedSet listeners; // of NullificationListener.

		private String host; // the name service host for the service.
		private String svcInterfaceName;
		private String svcProviderName;
		private String svcImplClassName;
		private LocalReference lref; // the service object.
		private File cacheDir;
	}

	/**
	 * New service objects are obtained using threads of this class.
	 * The important feature of this thread class is that it sets
	 * its context class loader to be the class loader specified
	 * in its "cl" argument. This way, RMI will obtain that class
	 * loader when it calls getContextClassLoader() on the thread,
	 * and RMI will then use that loader as the parent of any loaders
	 * it creates.
	 */
	private class LoaderThread extends Thread
	{
		private ClassLoader classLoader;
		private String svcInterfaceName;
		private String svcProviderName;
		private String host;
		private Requestor requestor;

		private Service service;
		private Throwable throwable;

		public LoaderThread
		(
			ClassLoader cl,
			String svcInterfaceName,
			String svcProviderName,
			String host, // name service host.
			Requestor requestor
		)
		{
			this.classLoader = cl;
			this.svcInterfaceName = svcInterfaceName;
			this.svcProviderName = svcProviderName;
			this.host = host;
			this.requestor = requestor;

			this.setContextClassLoader(cl);
		}

		public void run()
		{
			try
			{
				//
				// Check for valid input.

				if (host == null)
				{
					throw new Exception("Unspecified host.");
				}

				//
				// Obtain an updater, using the new class loader, and then connecting
				// to the remote update service for this object.

				Class remoteUpdaterClass = null;
				try
				{
					remoteUpdaterClass = classLoader.loadClass(RemoteUpdaterClassName);
				}
				catch (Exception ex)
				{
					throw new UpdateFailure(ex);
				}

				RemoteUpdater remoteUpdater = null;
				try
				{
					remoteUpdater = (RemoteUpdater)(remoteUpdaterClass.newInstance());
				}
				catch (Exception ex)
				{
					throw new UpdateFailure(ex);
				}

				//
				// Tell the updater which service to get and the name of
				// the service provider.

				remoteUpdater.setServiceInterfaceName(svcInterfaceName);
				remoteUpdater.setServiceProviderName(svcProviderName);

				//
				// Attempt to find the service provider.

				remoteUpdater.connect(host);

				//
				// Request a new instance.

				Service s = remoteUpdater.getNewInstance(requestor);
				System.out.println("Obtained instance: Version = " +
					s.getServiceVersionNo()[0]);

				//
				// Configure the new object, based on its local state.

				Configurator configurator = s.getConfigurator(MCRuntime.this);

				configurator.configure();
				System.out.println("configured");

				//
				// Return the new service object.

				service = s;
			}
			catch (Throwable t)
			{
				throwable = t;
			}
		}

		public Service getService()
		throws
			CallerIsCurrent,
			NoUpdateAvailable,
			ConfigurationAlert,
			ConfigurationFailure,
			ConnectionFailure,
			ServiceNotFound,
			UpdateFailure,
			InsufficientResources,
			RemoteException
		{
			if (throwable == null)
			{
				return service;
			}
			else
			{
				if (throwable instanceof CallerIsCurrent)
				{
					throw (CallerIsCurrent)throwable;
				}
				else if (throwable instanceof NoUpdateAvailable)
				{
					throw (NoUpdateAvailable)throwable;
				}
				else if (throwable instanceof ConfigurationAlert)
				{
					throw (ConfigurationAlert)throwable;
				}
				else if (throwable instanceof ConfigurationFailure)
				{
					throw (ConfigurationFailure)throwable;
				}
				else if (throwable instanceof ConnectionFailure)
				{
					throw (ConnectionFailure)throwable;
				}
				else if (throwable instanceof ServiceNotFound)
				{
					throw (ServiceNotFound)throwable;
				}
				else if (throwable instanceof UpdateFailure)
				{
					throw (UpdateFailure)throwable;
				}
				else if (throwable instanceof InsufficientResources)
				{
					throw (InsufficientResources)throwable;
				}
				else if (throwable instanceof Error)
				{
					throw (Error)throwable;
				}
				else
				{
					throw (RuntimeException)throwable;
				}
			}
		}
	}
}
