/*
 * Copyright (c) 2001 by Cliff Berg.
 * This code may be used and distributed for research for purposes.
 * No warrantee of any kind is given or implied.
 */

package mc.rmi;

import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.rmi.registry.LocateRegistry;
import java.rmi.NotBoundException;

/**
 * Finds a service provider, and allow a client to update a service.
 * Must have a zero-argument constructor, because this class is
 * loaded explicitly by a class loader and then instantiated
 * using Class.newInstance().
 */
public class RemoteUpdater
{
	/**
	 * Specify which service is to be updated. This is just the
	 * fully qualified name of the service class.
	 */
	public void setServiceInterfaceName(String svcInterfaceName)
	{
		this.svcInterfaceName = svcInterfaceName;
	}

	/**
	 * Specify the service provider name. This is the name specified
	 * to the name service (registry).
	 */
	public void setServiceProviderName(String svcProviderName)
	{
		this.svcProviderName = svcProviderName;
	}

	/**
	 * Attempt to locate the remote service provider. The host argument
	 * is the host on which the lookup (name) service should be found.
	 * This method retrieves the service provider stub if the service
	 * provider is found.
	 */
	public void connect(String host)
	throws
		RemoteException,
		ConnectionFailure
	{
		// Do an RMI lookup for the service.

		try
		{
			Registry r = LocateRegistry.getRegistry(host);
			if (r == null)
			{
				throw new ConnectionFailure(null, "RMI registry not found on " + host);
			}

			System.out.println("Looking up " + svcProviderName + " on " + host);
			rusvc = (ServiceProvider)(r.lookup(svcProviderName));
		}
		catch (Exception ex)
		{
			throw new ConnectionFailure(ex);
		}
	}

	public Service getNewInstance
	(
		Requestor requestor	// A callback object for the remote updater
			// servant to use to get any information it
			// needs. The requestor will likely implement
			// additional interfaces unique to the
			// specialized type of the object being
			// updated.
	)
	throws
		CallerIsCurrent,
		NoUpdateAvailable,
	 	ServiceNotFound,
		RemoteException
	{
		return rusvc.getNewInstance(svcInterfaceName, requestor);
	}

	private String svcInterfaceName;
	private String svcProviderName;
	private ServiceProvider rusvc;
}
